@@ -1,3 +1,7 @@
+"""Classes used to simulate recurly resources and endpoints
+
+Each endpoint class will define the CRUD interface into the resource.
+"""
 import recurly
 import six
 import random
@@ -7,28 +11,47 @@ import dateutil.parser
 
 from .utils import current_time
 from .errors import TRANSACTION_ERRORS, ResponseError
-from .core import details_route, serialize, serialize_list
+from .utils import details_route, serialize, serialize_list
 from .backend import accounts_backend, billing_info_backend, transactions_backend, invoices_backend, subscriptions_backend, plans_backend, plan_add_ons_backend, adjustments_backend, coupons_backend, coupon_redemptions_backend
 
+
 class BaseRecurlyEndpoint(object):
+    """Baseclass for simulating resource endpoints.
+
+    Provides basic CRUD functionality given a resource XML template, and object
+    store backend.
+    """
     pk_attr = 'uuid'
     XML = 0
     RAW = 1
 
     def hydrate_foreign_keys(self, obj):
+        """Hydrates all foreign key objects from Id strings into actual objects
+        """
         return obj
 
     def get_object_uri(self, obj):
+        """Returns the URI to access the given object resource
+        """
         cls = self.__class__
         return recurly.base_uri() + cls.base_uri + '/' + obj[cls.pk_attr]
 
     def uris(self, obj):
+        """Returns a dictionary of all URIs related to the object, including foreign keys
+        """
         obj = self.hydrate_foreign_keys(obj)
         uri_out = {}
         uri_out['object_uri'] = self.get_object_uri(obj)
         return uri_out
 
     def serialize(self, obj, format=XML):
+        """Serialize the object into the provided format, using the resource
+        template.
+
+        Currently only supports XML (for XML representation of the resource.
+        This is what recurly expects) and RAW (a dictionary representation of
+        the resource)
+        """
         if format == BaseRecurlyEndpoint.RAW:
             return obj
 
@@ -42,11 +65,15 @@ class BaseRecurlyEndpoint(object):
             return serialize(cls.template, cls.object_type, obj)
 
     def list(self, format=XML):
+        """Endpoint to list all resources stored in the backend
+        """
         cls = self.__class__
         out = cls.backend.list_objects()
         return self.serialize(out, format=format)
 
     def create(self, create_info, format=XML):
+        """Endpoint to create a new instance of the resource into the backend
+        """
         cls = self.__class__
         if cls.pk_attr in create_info:
             create_info['uuid'] = create_info[cls.pk_attr]
@@ -56,6 +83,10 @@ class BaseRecurlyEndpoint(object):
         return self.serialize(new_obj, format=format)
 
     def retrieve(self, pk, format=XML):
+        """Endpoint to retrieve an existing resource from the backend
+
+        Raises a 404 if the requested object does not exist.
+        """
         cls = self.__class__
         if not cls.backend.has_object(pk):
             raise ResponseError(404, '')
@@ -63,18 +94,33 @@ class BaseRecurlyEndpoint(object):
         return self.serialize(out, format=format)
 
     def update(self, pk, update_info, format=XML):
+        """Endpoint to update an existing resource from the backend
+
+        Raises a 404 if the requested object does not exist.
+        """
         cls = self.__class__
+        if not cls.backend.has_object(pk):
+            raise ResponseError(404, '')
         out = cls.backend.update_object(pk, update_info)
         return self.serialize(out, format=format)
 
     def delete(self, pk):
+        """Endpoint to delete an existing resource from the backend
+
+        Raises a 404 if the requested object does not exist.
+        """
         cls = self.__class__
+        if not cls.backend.has_object(pk):
+            raise ResponseError(404, '')
         cls.backend.delete_object(pk)
         return ''
 
     def generate_id(self):
+        """Generates a random ID that can be used as a UUID or recurly ID
+        """
         return ''.join(random.choice(string.ascii_lowercase + string.digits) for i in range(32))
 
+
 class AccountsEndpoint(BaseRecurlyEndpoint):
     base_uri = 'accounts'
     pk_attr = 'account_code'
@@ -120,6 +166,10 @@ class AccountsEndpoint(BaseRecurlyEndpoint):
         billing_info_backend.delete_object(pk)
         return ''
 
+    # Support for nested resources
+    # BillingInfo and CouponRedemption are managed by this endpoint, as
+    # opposed to having their own since Recurly API only provides access to these
+    # resources through the Account endpoint.
     def billing_info_uris(self, obj):
         uri_out = {}
         uri_out['account_uri'] = recurly.base_uri() + AccountsEndpoint.base_uri + '/' + obj['account']
@@ -192,6 +242,7 @@ class AccountsEndpoint(BaseRecurlyEndpoint):
         coupon_redemptions_backend.delete_object(coupon_redemption_uuid)
         return ''
 
+
 class TransactionsEndpoint(BaseRecurlyEndpoint):
     base_uri = 'transactions'
     backend = transactions_backend
@@ -204,9 +255,14 @@ class TransactionsEndpoint(BaseRecurlyEndpoint):
         return super(TransactionsEndpoint, self).__init__()
 
     def clear_state(self):
+        """Clears all registered errors
+        """
         self.registered_errors = {}
 
     def register_transaction_failure(self, account_code, error_code):
+        """Registers an error_code to associate with the given account for all
+        transactions made by the account
+        """
         self.registered_errors[account_code] = error_code
 
     def hydrate_foreign_keys(self, obj):
@@ -232,12 +288,14 @@ class TransactionsEndpoint(BaseRecurlyEndpoint):
         return uri_out
 
     def create(self, create_info, format=BaseRecurlyEndpoint.XML):
+        # Like recurly, creates an invoice that is associated with the
+        # transaction
         account_code = create_info['account'][AccountsEndpoint.pk_attr]
         assert AccountsEndpoint.backend.has_object(account_code)
         create_info['account'] = account_code
 
-        create_info['uuid'] = self.generate_id() # generate id now for invoice
-        create_info['tax_in_cents'] = 0 #unsupported
+        create_info['uuid'] = self.generate_id()  # generate id now for invoice
+        create_info['tax_in_cents'] = 0  # unsupported
         create_info['action'] = 'purchase'
         create_info['status'] = 'success'
         create_info['test'] = True
@@ -260,21 +318,19 @@ class TransactionsEndpoint(BaseRecurlyEndpoint):
             raise ResponseError(422, '<errors>{0}{1}</errors>'.format(error_xml, transaction_xml))
 
         # Every new transaction creates a new invoice
-        new_invoice = {
-                'account': account_code,
-                'uuid': self.generate_id(),
-                'state': 'collected',
-                'invoice_number': InvoicesEndpoint.generate_invoice_number(),
-                'subtotal_in_cents': int(create_info['amount_in_cents']),
-                'currency': create_info['currency'],
-                'created_at': create_info['created_at'],
-                'net_terms': 0,
-                'collection_method': 'automatic',
+        new_invoice = {'account': account_code,
+                       'uuid': self.generate_id(),
+                       'state': 'collected',
+                       'invoice_number': InvoicesEndpoint.generate_invoice_number(),
+                       'subtotal_in_cents': int(create_info['amount_in_cents']),
+                       'currency': create_info['currency'],
+                       'created_at': create_info['created_at'],
+                       'net_terms': 0,
+                       'collection_method': 'automatic',
 
-                # unsupported
-                'tax_type': 'usst',
-                'tax_rate': 0
-            }
+                       # unsupported
+                       'tax_type': 'usst',
+                       'tax_rate': 0}
         new_invoice['tax_in_cents'] = new_invoice['subtotal_in_cents'] * new_invoice['tax_rate']
         new_invoice['total_in_cents'] = new_invoice['subtotal_in_cents'] + new_invoice['tax_in_cents']
         new_invoice['transactions'] = [create_info['uuid']]
@@ -284,7 +340,9 @@ class TransactionsEndpoint(BaseRecurlyEndpoint):
         return super(TransactionsEndpoint, self).create(create_info, format)
 
     def delete(self, pk, amount_in_cents=None):
-        ''' DELETE is a refund action '''
+        """DELETE is a refund action, and as such this will not delete the
+        object from the backend.
+        """
         transaction = TransactionsEndpoint.backend.get_object(pk)
         if transaction['voidable'] and amount_in_cents is None:
             transaction['status'] = 'void'
@@ -300,29 +358,30 @@ class TransactionsEndpoint(BaseRecurlyEndpoint):
             if amount_in_cents is not None:
                 refund_transaction['amount_in_cents'] = amount_in_cents
             TransactionsEndpoint.backend.add_object(refund_transaction['uuid'], refund_transaction)
-            TransactionsEndpoint.backend.update_object(transaction['uuid'], {'refundable': False}) # Refunded, so now its no longer refundable
+            # Refunded, so now its no longer refundable
+            TransactionsEndpoint.backend.update_object(transaction['uuid'], {'refundable': False})
 
             invoice = InvoicesEndpoint.backend.get_object(transaction['invoice'])
-            InvoicesEndpoint.backend.update_object(transaction['invoice'], {'transactions': invoice['transactions']+[refund_transaction['uuid']]})
+            InvoicesEndpoint.backend.update_object(transaction['invoice'], {'transactions': invoice['transactions'] + [refund_transaction['uuid']]})
         else:
             # TODO: raise exception - transaction cannot be refunded
             pass
         return ''
 
+
 class AdjustmentsEndpoint(BaseRecurlyEndpoint):
     base_uri = 'adjustments'
     backend = adjustments_backend
     object_type = 'adjustment'
     object_type_plural = 'adjustments'
     template = 'adjustment.xml'
-    defaults = {
-            'state': 'active',
-            'quantity': 1,
-            'origin': 'credit',
-            'product_code': 'basic',
-            'discount_in_cents': 0,
-            'tax_exempt': False # unsupported
-        }
+    defaults = {'state': 'active',
+                'quantity': 1,
+                'origin': 'credit',
+                'product_code': 'basic',
+                'discount_in_cents': 0,
+                # unsupported
+                'tax_exempt': False}
 
     def uris(self, obj):
         uri_out = super(AdjustmentsEndpoint, self).uris(obj)
@@ -388,6 +447,7 @@ class InvoicesEndpoint(BaseRecurlyEndpoint):
             return '1000'
         return str(max(int(invoice['invoice_number']) for invoice in InvoicesEndpoint.backend.list_objects()) + 1)
 
+
 class CouponsEndpoint(BaseRecurlyEndpoint):
     base_uri = 'coupons'
     backend = coupons_backend
@@ -395,11 +455,9 @@ class CouponsEndpoint(BaseRecurlyEndpoint):
     object_type_plural = 'coupons'
     pk_attr = 'coupon_code'
     template = 'coupon.xml'
-    defaults = {
-            'state': 'redeemable',
-            'applies_to_all_plans': True,
-            'single_use': False
-        }
+    defaults = {'state': 'redeemable',
+                'applies_to_all_plans': True,
+                'single_use': False}
 
     def uris(self, obj):
         uri_out = super(CouponsEndpoint, self).uris(obj)
@@ -430,20 +488,22 @@ class CouponsEndpoint(BaseRecurlyEndpoint):
         return uri_out
 
     def serialize_coupon_redemption(self, obj, format=BaseRecurlyEndpoint.XML):
-        obj = self.hydrate_coupon_redemption_foreign_keys(obj)
         if format == BaseRecurlyEndpoint.RAW:
+            obj = self.hydrate_coupon_redemption_foreign_keys(obj)
             return obj
 
         if type(obj) == list:
+            obj = [self.hydrate_coupon_redemption_foreign_keys(o) for o in obj]
             for o in obj:
                 o['uris'] = self.coupon_redemption_uris(o)
             return serialize_list('redemption.xml', 'redemptions', 'redemption', obj)
         else:
+            obj = self.hydrate_coupon_redemption_foreign_keys(obj)
             obj['uris'] = self.coupon_redemption_uris(obj)
             return serialize('redemption.xml', 'redemption', obj)
 
     @details_route('GET', 'redemptions', is_list=True)
-    def get_coupon_redemptions(self, pk, format=BaseRecurlyEndpoint.XML):
+    def get_coupon_redemptions(self, pk, filters=None, format=BaseRecurlyEndpoint.XML):
         obj_list = coupon_redemptions_backend.list_objects(lambda redemption: redemption['coupon'] == pk)
         return self.serialize_coupon_redemption(obj_list, format=format)
 
@@ -461,6 +521,7 @@ class CouponsEndpoint(BaseRecurlyEndpoint):
         else:
             return int(coupon['discount_in_cents'])
 
+
 class PlansEndpoint(BaseRecurlyEndpoint):
     base_uri = 'plans'
     backend = plans_backend
@@ -468,18 +529,15 @@ class PlansEndpoint(BaseRecurlyEndpoint):
     object_type = 'plan'
     object_type_plural = 'plans'
     template = 'plan.xml'
-    defaults = {
-        'plan_interval_unit': 'months',
-        'plan_interval_length': 1,
-        'trial_interval_unit': 'months',
-        'trial_interval_length': 0,
-        'display_quantity': False,
-        'tax_exempt': False # unsupported
-    }
-    add_on_defaults = {
-        'default_quantity': 1,
-        'display_quantity_on_hosted_page': False
-    }
+    defaults = {'plan_interval_unit': 'months',
+                'plan_interval_length': 1,
+                'trial_interval_unit': 'months',
+                'trial_interval_length': 0,
+                'display_quantity': False,
+                # unsupported
+                'tax_exempt': False}
+    add_on_defaults = {'default_quantity': 1,
+                       'display_quantity_on_hosted_page': False}
 
     def uris(self, obj):
         uri_out = super(PlansEndpoint, self).uris(obj)
@@ -529,13 +587,14 @@ class PlansEndpoint(BaseRecurlyEndpoint):
             create_info['accounting_code'] = create_info['add_on_code']
         return self.serialize_plan_add_on(plan_add_ons_backend.add_object(self.generate_plan_add_on_uuid(pk, create_info['add_on_code']), create_info), format=format)
 
+
 class SubscriptionsEndpoint(BaseRecurlyEndpoint):
     base_uri = 'subscriptions'
     backend = subscriptions_backend
     object_type = 'subscription'
     object_type_plural = 'subscriptions'
     template = 'subscription.xml'
-    defaults = { 'quantity': 1, 'collection_method': 'automatic' }
+    defaults = {'quantity': 1, 'collection_method': 'automatic'}
 
     def _calculate_timedelta(self, units, length):
         timedelta_info = {}
@@ -572,6 +631,8 @@ class SubscriptionsEndpoint(BaseRecurlyEndpoint):
         return uri_out
 
     def create(self, create_info, format=BaseRecurlyEndpoint.XML):
+        # Like recurly, this will create a new invoice and transaction that
+        # goes with the new subscription enrollment
         account_code = create_info['account'][AccountsEndpoint.pk_attr]
         if not AccountsEndpoint.backend.has_object(account_code):
             accounts_endpoint.create(create_info['account'])
@@ -595,7 +656,8 @@ class SubscriptionsEndpoint(BaseRecurlyEndpoint):
         if 'starts_at' in create_info:
             # A custom start date is specified
             create_info['activated_at'] = create_info['starts_at']
-            create_info['current_period_started_at'] = create_info['starts_at'] # TODO: confirm recurly sets current_period_started_at for future subs
+            # TODO: confirm recurly sets current_period_started_at for future subs
+            create_info['current_period_started_at'] = create_info['starts_at']
         elif 'trial_started_at' in create_info:
             create_info['activated_at'] = self._parse_isoformat(create_info['trial_ends_at'])
             create_info['current_period_started_at'] = create_info['trial_started_at']
@@ -605,7 +667,8 @@ class SubscriptionsEndpoint(BaseRecurlyEndpoint):
             create_info['current_period_started_at'] = now.isoformat()
 
         started_at = self._parse_isoformat(create_info['current_period_started_at'])
-        if now >= started_at: # Plan already started
+        if now >= started_at:
+            # Plan already started
             if 'first_renewal_date' in create_info:
                 create_info['current_period_ends_at'] = self._parse_isoformat(create_info['first_renewal_date'])
             else:
@@ -642,28 +705,29 @@ class SubscriptionsEndpoint(BaseRecurlyEndpoint):
         self.hydrate_foreign_keys(new_sub)
 
         if defaults['state'] == 'active':
-            # Setup charges first, to calculate total charge to put on the invoice and transaction
+            # Setup charges first, to calculate total charge to put on the
+            # invoice and transaction
             total = 0
             adjustment_infos = []
             plan_charge_line_item = {
-                        'account_code': new_sub['account'],
-                        'currency': new_sub['currency'],
-                        'unit_amount_in_cents': int(new_sub['unit_amount_in_cents']),
-                        'description': new_sub['plan']['name'],
-                        'quantity': new_sub['quantity']
-                    }
+                'account_code': new_sub['account'],
+                'currency': new_sub['currency'],
+                'unit_amount_in_cents': int(new_sub['unit_amount_in_cents']),
+                'description': new_sub['plan']['name'],
+                'quantity': new_sub['quantity']
+            }
             total += plan_charge_line_item['unit_amount_in_cents']
             adjustment_infos.append(plan_charge_line_item)
 
             if 'subscription_add_ons' in new_sub:
                 for add_on in new_sub['subscription_add_ons']:
                     plan_charge_line_item = {
-                                'account_code': new_sub['account'],
-                                'currency': new_sub['currency'],
-                                'unit_amount_in_cents': int(add_on['unit_amount_in_cents']),
-                                'description': add_on['name'],
-                                'quantity': new_sub['quantity'],
-                            }
+                        'account_code': new_sub['account'],
+                        'currency': new_sub['currency'],
+                        'unit_amount_in_cents': int(add_on['unit_amount_in_cents']),
+                        'description': add_on['name'],
+                        'quantity': new_sub['quantity'],
+                    }
                     total += plan_charge_line_item['unit_amount_in_cents']
                     adjustment_infos.append(plan_charge_line_item)
 
@@ -717,19 +781,19 @@ class SubscriptionsEndpoint(BaseRecurlyEndpoint):
             transactions_endpoint.delete(transaction[TransactionsEndpoint.pk_attr])
 
         return self.serialize(SubscriptionsEndpoint.backend.update_object(pk, {
-                'state': 'expired',
-                'expires_at': now.isoformat(),
-                'current_period_ends_at': now.isoformat()
-            }), format=format)
+            'state': 'expired',
+            'expires_at': now.isoformat(),
+            'current_period_ends_at': now.isoformat()
+        }), format=format)
 
     @details_route('PUT', 'cancel')
     def cancel_subscription(self, pk, cancel_info, format=format):
         subscription = SubscriptionsEndpoint.backend.get_object(pk)
         return self.serialize(SubscriptionsEndpoint.backend.update_object(pk, {
-                'state': 'canceled',
-                'expires_at': subscription['current_period_ends_at'],
-                'canceled_at': current_time().isoformat()
-            }), format=format)
+            'state': 'canceled',
+            'expires_at': subscription['current_period_ends_at'],
+            'canceled_at': current_time().isoformat()
+        }), format=format)
 
 accounts_endpoint = AccountsEndpoint()
 adjustments_endpoint = AdjustmentsEndpoint()
@@ -739,12 +803,16 @@ invoices_endpoint = InvoicesEndpoint()
 plans_endpoint = PlansEndpoint()
 subscriptions_endpoint = SubscriptionsEndpoint()
 endpoints = [accounts_endpoint,
-        adjustments_endpoint,
-        transactions_endpoint,
-        coupons_endpoint,
-        invoices_endpoint,
-        plans_endpoint,
-        subscriptions_endpoint]
+             adjustments_endpoint,
+             transactions_endpoint,
+             coupons_endpoint,
+             invoices_endpoint,
+             plans_endpoint,
+             subscriptions_endpoint]
+
 
 def clear_endpoints():
+    """Clear state off of all endpoints. This ensures that no residual state
+    carries over between mocurly contexts.
+    """
     transactions_endpoint.clear_state()
